@@ -1,5 +1,24 @@
 /* User-kernel network link */
 
+/*
+ * This file is part of secnet.
+ * See README for full list of copyright holders.
+ *
+ * secnet is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version d of the License, or
+ * (at your option) any later version.
+ * 
+ * secnet is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * version 3 along with secnet; if not, see
+ * https://www.gnu.org/licenses/gpl.html.
+ */
+
 /* See RFCs 791, 792, 1123 and 1812 */
 
 /* The netlink device is actually a router.  Tunnels are unnumbered
@@ -106,6 +125,12 @@ their use.
 #include "netlink.h"
 #include "process.h"
 
+#ifdef NETLINK_DEBUG
+#define MDEBUG(...) Message(M_DEBUG, __VA_ARGS__)
+#else /* !NETLINK_DEBUG */
+#define MDEBUG(...) ((void)0)
+#endif /* !NETLINK_DEBUG */
+
 #define ICMP_TYPE_ECHO_REPLY             0
 
 #define ICMP_TYPE_UNREACHABLE            3
@@ -120,7 +145,7 @@ their use.
 #define ICMP_CODE_TTL_EXCEEDED           0
 
 /* Generic IP checksum routine */
-static inline uint16_t ip_csum(uint8_t *iph,int32_t count)
+static inline uint16_t ip_csum(const uint8_t *iph,int32_t count)
 {
     register uint32_t sum=0;
 
@@ -144,7 +169,7 @@ static inline uint16_t ip_csum(uint8_t *iph,int32_t count)
  *      By Jorge Cwik <jorge@laser.satlink.net>, adapted for linux by
  *      Arnt Gulbrandsen.
  */
-static inline uint16_t ip_fast_csum(uint8_t *iph, int32_t ihl) {
+static inline uint16_t ip_fast_csum(const uint8_t *iph, int32_t ihl) {
     uint32_t sum;
 
     __asm__ __volatile__(
@@ -174,7 +199,7 @@ static inline uint16_t ip_fast_csum(uint8_t *iph, int32_t ihl) {
     return sum;
 }
 #else
-static inline uint16_t ip_fast_csum(uint8_t *iph, int32_t ihl)
+static inline uint16_t ip_fast_csum(const uint8_t *iph, int32_t ihl)
 {
     assert(ihl < INT_MAX/4);
     return ip_csum(iph,ihl*4);
@@ -192,7 +217,11 @@ struct iphdr {
     uint8_t    tos;
     uint16_t   tot_len;
     uint16_t   id;
-    uint16_t   frag_off;
+    uint16_t   frag;
+#define IPHDR_FRAG_OFF  ((uint16_t)0x1fff)
+#define IPHDR_FRAG_MORE ((uint16_t)0x2000)
+#define IPHDR_FRAG_DONT ((uint16_t)0x4000)
+/*                 reserved        0x8000 */
     uint8_t    ttl;
     uint8_t    protocol;
     uint16_t   check;
@@ -206,7 +235,7 @@ struct icmphdr {
     uint8_t type;
     uint8_t code;
     uint16_t check;
-    union {
+    union icmpinfofield {
 	uint32_t unused;
 	struct {
 	    uint8_t pointer;
@@ -218,9 +247,29 @@ struct icmphdr {
 	    uint16_t id;
 	    uint16_t seq;
 	} echo;
+	struct {
+	    uint16_t unused;
+	    uint16_t mtu;
+	} fragneeded;
     } d;
 };
+
+static const union icmpinfofield icmp_noinfo;
     
+static void netlink_client_deliver(struct netlink *st,
+				   struct netlink_client *client,
+				   uint32_t source, uint32_t dest,
+				   struct buffer_if *buf);
+static void netlink_host_deliver(struct netlink *st,
+				 struct netlink_client *sender,
+				 uint32_t source, uint32_t dest,
+				 struct buffer_if *buf);
+
+static const char *sender_name(struct netlink_client *sender /* or NULL */)
+{
+    return sender?sender->name:"(local)";
+}
+
 static void netlink_packet_deliver(struct netlink *st,
 				   struct netlink_client *client,
 				   struct buffer_if *buf);
@@ -233,7 +282,8 @@ static void netlink_packet_deliver(struct netlink *st,
    settable.
    */
 static struct icmphdr *netlink_icmp_tmpl(struct netlink *st,
-					 uint32_t dest,uint16_t len)
+					 uint32_t source, uint32_t dest,
+					 uint16_t len)
 {
     struct icmphdr *h;
 
@@ -246,10 +296,10 @@ static struct icmphdr *netlink_icmp_tmpl(struct netlink *st,
     h->iph.tos=0;
     h->iph.tot_len=htons(len+(h->iph.ihl*4)+8);
     h->iph.id=0;
-    h->iph.frag_off=0;
+    h->iph.frag=0;
     h->iph.ttl=255; /* XXX should be configurable */
     h->iph.protocol=1;
-    h->iph.saddr=htonl(st->secnet_address);
+    h->iph.saddr=htonl(source);
     h->iph.daddr=htonl(dest);
     h->iph.check=0;
     h->iph.check=ip_fast_csum((uint8_t *)&h->iph,h->iph.ihl);
@@ -298,14 +348,22 @@ static bool_t netlink_icmp_may_reply(struct buffer_if *buf)
     icmph=(struct icmphdr *)buf->start;
     if (iph->protocol==1) {
 	switch(icmph->type) {
-	case 3: /* Destination unreachable */
-	case 11: /* Time Exceeded */
-	case 12: /* Parameter Problem */
+	    /* Based on http://www.iana.org/assignments/icmp-parameters/icmp-parameters.xhtml#icmp-parameters-types
+	     * as retrieved Thu, 20 Mar 2014 00:16:44 +0000.
+	     * Deprecated, reserved, unassigned and experimental
+	     * options are treated as not safe to reply to.
+	     */
+	case 0: /* Echo Reply */
+	case 8: /* Echo */
+	case 13: /* Timestamp */
+	case 14: /* Timestamp Reply */
+	    return True;
+	default:
 	    return False;
 	}
     }
     /* How do we spot broadcast destination addresses? */
-    if (ntohs(iph->frag_off)&0x1fff) return False; /* Non-initial fragment */
+    if (ntohs(iph->frag)&IPHDR_FRAG_OFF) return False;
     source=ntohl(iph->saddr);
     if (source==0) return False;
     if ((source&0xff000000)==0x7f000000) return False;
@@ -347,26 +405,60 @@ static uint16_t netlink_icmp_reply_len(struct buffer_if *buf)
     /* We include the first 8 bytes of the packet data, provided they exist */
     hlen+=8;
     plen=ntohs(iph->tot_len);
-    return (hlen>plen?plen:hlen);
+    return MIN(hlen,plen);
 }
 
 /* client indicates where the packet we're constructing a response to
    comes from. NULL indicates the host. */
-static void netlink_icmp_simple(struct netlink *st, struct buffer_if *buf,
-				struct netlink_client *client,
-				uint8_t type, uint8_t code)
+static void netlink_icmp_simple(struct netlink *st,
+				struct netlink_client *origsender,
+				struct buffer_if *buf,
+				uint8_t type, uint8_t code,
+				union icmpinfofield info)
 {
     struct icmphdr *h;
     uint16_t len;
 
     if (netlink_icmp_may_reply(buf)) {
 	struct iphdr *iph=(struct iphdr *)buf->start;
+
+	uint32_t icmpdest = ntohl(iph->saddr);
+	uint32_t icmpsource;
+	const char *icmpsourcedebugprefix;
+	if (!st->ptp) {
+	    icmpsource=st->secnet_address;
+	    icmpsourcedebugprefix="";
+	} else if (origsender) {
+	    /* was from peer, send reply as if from host */
+	    icmpsource=st->local_address;
+	    icmpsourcedebugprefix="L!";
+	} else {
+	    /* was from host, send reply as if from peer */
+	    icmpsource=st->secnet_address; /* actually, peer address */
+	    icmpsourcedebugprefix="P!";
+	}
+	MDEBUG("%s: generating ICMP re %s[%s]->[%s]:"
+	       " from %s%s type=%u code=%u\n",
+	       st->name, sender_name(origsender),
+	       ipaddr_to_string(ntohl(iph->saddr)),
+	       ipaddr_to_string(ntohl(iph->daddr)),
+	       icmpsourcedebugprefix,
+	       ipaddr_to_string(icmpsource),
+	       type, code);
+
 	len=netlink_icmp_reply_len(buf);
-	h=netlink_icmp_tmpl(st,ntohl(iph->saddr),len);
-	h->type=type; h->code=code;
-	memcpy(buf_append(&st->icmp,len),buf->start,len);
+	h=netlink_icmp_tmpl(st,icmpsource,icmpdest,len);
+	h->type=type; h->code=code; h->d=info;
+	BUF_ADD_BYTES(append,&st->icmp,buf->start,len);
 	netlink_icmp_csum(h);
-	netlink_packet_deliver(st,NULL,&st->icmp);
+
+	if (!st->ptp) {
+	    netlink_packet_deliver(st,NULL,&st->icmp);
+	} else if (origsender) {
+	    netlink_client_deliver(st,origsender,icmpsource,icmpdest,&st->icmp);
+	} else {
+	    netlink_host_deliver(st,NULL,icmpsource,icmpdest,&st->icmp);
+	}
 	BUF_ASSERT_FREE(&st->icmp);
     }
 }
@@ -409,16 +501,190 @@ static bool_t netlink_check(struct netlink *st, struct buffer_if *buf,
 #undef BAD
 }
 
-/* Deliver a packet. "client" is the _origin_ of the packet, not its
-   destination, and is NULL for packets from the host and packets
+static const char *fragment_filter_header(uint8_t *base, long *hlp)
+{
+    const int fixedhl = sizeof(struct iphdr);
+    long hl = *hlp;
+    const uint8_t *ipend = base + hl;
+    uint8_t *op = base + fixedhl;
+    const uint8_t *ip = op;
+
+    while (ip < ipend) {
+	uint8_t opt = ip[0];
+	int remain = ipend - ip;
+	if (opt == 0x00) /* End of Options List */ break;
+	if (opt == 0x01) /* No Operation */ continue;
+	if (remain < 2) return "IPv4 options truncated at length";
+	int optlen = ip[1];
+	if (remain < optlen) return "IPv4 options truncated in option";
+	if (opt & 0x80) /* copy */ {
+	    memmove(op, ip, optlen);
+	    op += optlen;
+	}
+	ip += optlen;
+    }
+    while ((hl = (op - base)) & 0x3)
+	*op++ = 0x00 /* End of Option List */;
+    ((struct iphdr*)base)->ihl = hl >> 2;
+    *hlp = hl;
+
+    return 0;
+}
+
+/* Fragment or send ICMP Fragmentation Needed */
+static void netlink_maybe_fragment(struct netlink *st,
+				   struct netlink_client *sender,
+				   netlink_deliver_fn *deliver,
+				   void *deliver_dst,
+				   const char *delivery_name,
+				   int32_t mtu,
+				   uint32_t source, uint32_t dest,
+				   struct buffer_if *buf)
+{
+    struct iphdr *iph=(struct iphdr*)buf->start;
+    long hl = iph->ihl*4;
+    const char *ssource = ipaddr_to_string(source);
+
+    if (buf->size <= mtu) {
+	deliver(deliver_dst, buf);
+	return;
+    }
+
+    MDEBUG("%s: fragmenting %s->%s org.size=%"PRId32"\n",
+	   st->name, ssource, delivery_name, buf->size);
+
+#define BADFRAG(m, ...)					\
+	Message(M_WARNING,				\
+	        "%s: fragmenting packet from source %s"	\
+		" for transmission via %s: " m "\n",	\
+		st->name, ssource, delivery_name,	\
+		## __VA_ARGS__);
+
+    unsigned orig_frag = ntohs(iph->frag);
+
+    if (orig_frag&IPHDR_FRAG_DONT) {
+	union icmpinfofield info =
+	    { .fragneeded = { .unused = 0, .mtu = htons(mtu) } };
+	netlink_icmp_simple(st,sender,buf,
+			    ICMP_TYPE_UNREACHABLE,
+			    ICMP_CODE_FRAGMENTATION_REQUIRED,
+			    info);
+	BUF_FREE(buf);
+	return;
+    }
+    if (mtu < hl + 8) {
+	BADFRAG("mtu %"PRId32" too small", mtu);
+	BUF_FREE(buf);
+	return;
+    }
+
+    /* we (ab)use the icmp buffer to stash the original packet */
+    struct buffer_if *orig = &st->icmp;
+    BUF_ALLOC(orig,"netlink_client_deliver fragment orig");
+    buffer_copy(orig,buf);
+    BUF_FREE(buf);
+
+    const uint8_t *startindata = orig->start + hl;
+    const uint8_t *indata =      startindata;
+    const uint8_t *endindata =   orig->start + orig->size;
+    _Bool filtered = 0;
+
+    for (;;) {
+	/* compute our fragment offset */
+	long dataoffset = indata - startindata
+	    + (orig_frag & IPHDR_FRAG_OFF)*8;
+	assert(!(dataoffset & 7));
+	if (dataoffset > IPHDR_FRAG_OFF*8) {
+	    BADFRAG("ultimate fragment offset out of range");
+	    break;
+	}
+
+	BUF_ALLOC(buf,"netlink_client_deliver fragment frag");
+	buffer_init(buf,calculate_max_start_pad());
+
+	/* copy header (possibly filtered); will adjust in a bit */
+	struct iphdr *fragh = buf_append(buf, hl);
+	memcpy(fragh, orig->start, hl);
+
+	/* decide how much payload to copy and copy it */
+	long avail = mtu - hl;
+	long remain = endindata - indata;
+	long use = avail < remain ? (avail & ~(long)7) : remain;
+	BUF_ADD_BYTES(append, buf, indata, use);
+	indata += use;
+
+	_Bool last_frag = indata >= endindata;
+
+	/* adjust the header */
+	fragh->tot_len = htons(buf->size);
+	fragh->frag =
+	    htons((orig_frag & ~IPHDR_FRAG_OFF) |
+		  (last_frag ? 0 : IPHDR_FRAG_MORE) |
+		  (dataoffset >> 3));
+	fragh->check = 0;
+	fragh->check = ip_fast_csum((const void*)fragh, fragh->ihl);
+
+	/* actually send it */
+	deliver(deliver_dst, buf);
+	if (last_frag)
+	    break;
+
+	/* after copying the header for the first frag,
+	 * we filter the header for the remaining frags */
+	if (!filtered++) {
+	    const char *bad = fragment_filter_header(orig->start, &hl);
+	    if (bad) { BADFRAG("%s", bad); break; }
+	}
+    }
+
+    BUF_FREE(orig);
+
+#undef BADFRAG
+}
+
+/* Deliver a packet _to_ client; used after we have decided
+ * what to do with it (and just to check that the client has
+ * actually registered a delivery function with us). */
+static void netlink_client_deliver(struct netlink *st,
+				   struct netlink_client *client,
+				   uint32_t source, uint32_t dest,
+				   struct buffer_if *buf)
+{
+    if (!client->deliver) {
+	string_t s,d;
+	s=ipaddr_to_string(source);
+	d=ipaddr_to_string(dest);
+	Message(M_ERR,"%s: dropping %s->%s, client not registered\n",
+		st->name,s,d);
+	BUF_FREE(buf);
+	return;
+    }
+    netlink_maybe_fragment(st,NULL, client->deliver,client->dst,client->name,
+			   client->mtu, source,dest,buf);
+    client->outcount++;
+}
+
+/* Deliver a packet to the host; used after we have decided that that
+ * is what to do with it. */
+static void netlink_host_deliver(struct netlink *st,
+				 struct netlink_client *sender,
+				 uint32_t source, uint32_t dest,
+				 struct buffer_if *buf)
+{
+    netlink_maybe_fragment(st,sender, st->deliver_to_host,st->dst,"(host)",
+			   st->mtu, source,dest,buf);
+    st->outcount++;
+}
+
+/* Deliver a packet. "sender"==NULL for packets from the host and packets
    generated internally in secnet.  */
 static void netlink_packet_deliver(struct netlink *st,
-				   struct netlink_client *client,
+				   struct netlink_client *sender,
 				   struct buffer_if *buf)
 {
     if (buf->size < (int)sizeof(struct iphdr)) {
 	Message(M_ERR,"%s: trying to deliver a too-short packet"
-		" from %s!\n",st->name, client?client->name:"(local)");
+		" from %s!\n",st->name, sender_name(sender));
 	BUF_FREE(buf);
 	return;
     }
@@ -440,9 +706,9 @@ static void netlink_packet_deliver(struct netlink *st,
 	return;
     }
     
-    /* Packets from the host (client==NULL) may always be routed.  Packets
+    /* Packets from the host (sender==NULL) may always be routed.  Packets
        from clients with the allow_route option will also be routed. */
-    if (!client || (client && (client->options & OPT_ALLOWROUTE)))
+    if (!sender || (sender && (sender->options & OPT_ALLOWROUTE)))
 	allow_route=True;
 
     /* If !allow_route, we check the routing table anyway, and if
@@ -488,8 +754,7 @@ static void netlink_packet_deliver(struct netlink *st,
 	/* The packet's not going down a tunnel.  It might (ought to)
 	   be for the host.   */
 	if (ipset_contains_addr(st->networks,dest)) {
-	    st->deliver_to_host(st->dst,buf);
-	    st->outcount++;
+	    netlink_host_deliver(st,sender,source,dest,buf);
 	    BUF_ASSERT_FREE(buf);
 	} else {
 	    string_t s,d;
@@ -497,9 +762,8 @@ static void netlink_packet_deliver(struct netlink *st,
 	    d=ipaddr_to_string(dest);
 	    Message(M_DEBUG,"%s: don't know where to deliver packet "
 		    "(s=%s, d=%s)\n", st->name, s, d);
-	    free(s); free(d);
-	    netlink_icmp_simple(st,buf,client,ICMP_TYPE_UNREACHABLE,
-				ICMP_CODE_NET_UNREACHABLE);
+	    netlink_icmp_simple(st,sender,buf,ICMP_TYPE_UNREACHABLE,
+				ICMP_CODE_NET_UNREACHABLE, icmp_noinfo);
 	    BUF_FREE(buf);
 	}
     } else {
@@ -513,22 +777,21 @@ static void netlink_packet_deliver(struct netlink *st,
 	       with destination network administratively prohibited */
 	    Message(M_NOTICE,"%s: denied forwarding for packet (s=%s, d=%s)\n",
 		    st->name,s,d);
-	    free(s); free(d);
 		    
-	    netlink_icmp_simple(st,buf,client,ICMP_TYPE_UNREACHABLE,
-				ICMP_CODE_NET_PROHIBITED);
+	    netlink_icmp_simple(st,sender,buf,ICMP_TYPE_UNREACHABLE,
+				ICMP_CODE_NET_PROHIBITED, icmp_noinfo);
 	    BUF_FREE(buf);
 	} else {
 	    if (best_quality>0) {
-		/* XXX Fragment if required */
-		st->routes[best_match]->deliver(
-		    st->routes[best_match]->dst, buf);
-		st->routes[best_match]->outcount++;
+		netlink_client_deliver(st,st->routes[best_match],
+				       source,dest,buf);
 		BUF_ASSERT_FREE(buf);
 	    } else {
 		/* Generate ICMP destination unreachable */
-		netlink_icmp_simple(st,buf,client,ICMP_TYPE_UNREACHABLE,
-				    ICMP_CODE_NET_UNREACHABLE); /* client==NULL */
+		netlink_icmp_simple(st,sender,buf,
+				    ICMP_TYPE_UNREACHABLE,
+				    ICMP_CODE_NET_UNREACHABLE,
+				    icmp_noinfo);
 		BUF_FREE(buf);
 	    }
 	}
@@ -537,7 +800,7 @@ static void netlink_packet_deliver(struct netlink *st,
 }
 
 static void netlink_packet_forward(struct netlink *st, 
-				   struct netlink_client *client,
+				   struct netlink_client *sender,
 				   struct buffer_if *buf)
 {
     if (buf->size < (int)sizeof(struct iphdr)) return;
@@ -548,8 +811,8 @@ static void netlink_packet_forward(struct netlink *st,
     /* Packet has already been checked */
     if (iph->ttl<=1) {
 	/* Generate ICMP time exceeded */
-	netlink_icmp_simple(st,buf,client,ICMP_TYPE_TIME_EXCEEDED,
-			    ICMP_CODE_TTL_EXCEEDED);
+	netlink_icmp_simple(st,sender,buf,ICMP_TYPE_TIME_EXCEEDED,
+			    ICMP_CODE_TTL_EXCEEDED,icmp_noinfo);
 	BUF_FREE(buf);
 	return;
     }
@@ -557,13 +820,13 @@ static void netlink_packet_forward(struct netlink *st,
     iph->check=0;
     iph->check=ip_fast_csum((uint8_t *)iph,iph->ihl);
 
-    netlink_packet_deliver(st,client,buf);
+    netlink_packet_deliver(st,sender,buf);
     BUF_ASSERT_FREE(buf);
 }
 
 /* Deal with packets addressed explicitly to us */
 static void netlink_packet_local(struct netlink *st,
-				 struct netlink_client *client,
+				 struct netlink_client *sender,
 				 struct buffer_if *buf)
 {
     struct icmphdr *h;
@@ -578,9 +841,12 @@ static void netlink_packet_local(struct netlink *st,
     }
     h=(struct icmphdr *)buf->start;
 
-    if ((ntohs(h->iph.frag_off)&0xbfff)!=0) {
-	Message(M_WARNING,"%s: fragmented packet addressed to secnet; "
-		"ignoring it\n",st->name);
+    unsigned fraginfo = ntohs(h->iph.frag);
+    if ((fraginfo&(IPHDR_FRAG_OFF|IPHDR_FRAG_MORE))!=0) {
+	if (!(fraginfo & IPHDR_FRAG_OFF))
+	    /* report only for first fragment */
+	    Message(M_WARNING,"%s: fragmented packet addressed to secnet; "
+		    "ignoring it\n",st->name);
 	BUF_FREE(buf);
 	return;
     }
@@ -603,8 +869,8 @@ static void netlink_packet_local(struct netlink *st,
 	Message(M_WARNING,"%s: unknown incoming ICMP\n",st->name);
     } else {
 	/* Send ICMP protocol unreachable */
-	netlink_icmp_simple(st,buf,client,ICMP_TYPE_UNREACHABLE,
-			    ICMP_CODE_PROTOCOL_UNREACHABLE);
+	netlink_icmp_simple(st,sender,buf,ICMP_TYPE_UNREACHABLE,
+			    ICMP_CODE_PROTOCOL_UNREACHABLE,icmp_noinfo);
 	BUF_FREE(buf);
 	return;
     }
@@ -614,13 +880,13 @@ static void netlink_packet_local(struct netlink *st,
 
 /* If cid==NULL packet is from host, otherwise cid specifies which tunnel 
    it came from. */
-static void netlink_incoming(struct netlink *st, struct netlink_client *client,
+static void netlink_incoming(struct netlink *st, struct netlink_client *sender,
 			     struct buffer_if *buf)
 {
     uint32_t source,dest;
     struct iphdr *iph;
     char errmsgbuf[50];
-    const char *sourcedesc=client?client->name:"host";
+    const char *sourcedesc=sender?sender->name:"host";
 
     BUF_ASSERT_USED(buf);
 
@@ -631,7 +897,7 @@ static void netlink_incoming(struct netlink *st, struct netlink_client *client,
 	BUF_FREE(buf);
 	return;
     }
-    assert(buf->size >= (int)sizeof(struct icmphdr));
+    assert(buf->size >= (int)sizeof(struct iphdr));
     iph=(struct iphdr *)buf->start;
 
     source=ntohl(iph->saddr);
@@ -640,16 +906,15 @@ static void netlink_incoming(struct netlink *st, struct netlink_client *client,
     /* Check source. If we don't like the source, there's no point
        generating ICMP because we won't know how to get it to the
        source of the packet. */
-    if (client) {
+    if (sender) {
 	/* Check that the packet source is appropriate for the tunnel
 	   it came down */
-	if (!ipset_contains_addr(client->networks,source)) {
+	if (!ipset_contains_addr(sender->networks,source)) {
 	    string_t s,d;
 	    s=ipaddr_to_string(source);
 	    d=ipaddr_to_string(dest);
 	    Message(M_WARNING,"%s: packet from tunnel %s with bad "
-		    "source address (s=%s,d=%s)\n",st->name,client->name,s,d);
-	    free(s); free(d);
+		    "source address (s=%s,d=%s)\n",st->name,sender->name,s,d);
 	    BUF_FREE(buf);
 	    return;
 	}
@@ -663,7 +928,6 @@ static void netlink_incoming(struct netlink *st, struct netlink_client *client,
 	    d=ipaddr_to_string(dest);
 	    Message(M_WARNING,"%s: outgoing packet with bad source address "
 		    "(s=%s,d=%s)\n",st->name,s,d);
-	    free(s); free(d);
 	    BUF_FREE(buf);
 	    return;
 	}
@@ -675,10 +939,10 @@ static void netlink_incoming(struct netlink *st, struct netlink_client *client,
        where it came from.  It's up to external software to check
        address validity and generate ICMP, etc. */
     if (st->ptp) {
-	if (client) {
-	    st->deliver_to_host(st->dst,buf);
+	if (sender) {
+	    netlink_host_deliver(st,sender,source,dest,buf);
 	} else {
-	    st->clients->deliver(st->clients->dst,buf);
+	    netlink_client_deliver(st,st->clients,source,dest,buf);
 	}
 	BUF_ASSERT_FREE(buf);
 	return;
@@ -687,11 +951,11 @@ static void netlink_incoming(struct netlink *st, struct netlink_client *client,
     /* st->secnet_address needs checking before matching destination
        addresses */
     if (dest==st->secnet_address) {
-	netlink_packet_local(st,client,buf);
+	netlink_packet_local(st,sender,buf);
 	BUF_ASSERT_FREE(buf);
 	return;
     }
-    netlink_packet_forward(st,client,buf);
+    netlink_packet_forward(st,sender,buf);
     BUF_ASSERT_FREE(buf);
 }
 
@@ -731,7 +995,6 @@ static void netlink_output_subnets(struct netlink *st, uint32_t loglevel,
     for (i=0; i<snets->entries; i++) {
 	net=subnet_to_string(snets->list[i]);
 	Message(loglevel,"%s ",net);
-	free(net);
     }
 }
 
@@ -744,9 +1007,8 @@ static void netlink_dump_routes(struct netlink *st, bool_t requested)
     if (requested) c=M_WARNING;
     if (st->ptp) {
 	net=ipaddr_to_string(st->secnet_address);
-	Message(c,"%s: point-to-point (remote end is %s); routes:\n",
+	Message(c,"%s: point-to-point (remote end is %s); routes: ",
 		st->name, net);
-	free(net);
 	netlink_output_subnets(st,c,st->clients->subnets);
 	Message(c,"\n");
     } else {
@@ -767,11 +1029,9 @@ static void netlink_dump_routes(struct netlink *st, bool_t requested)
 	net=ipaddr_to_string(st->secnet_address);
 	Message(c,"%s/32 -> netlink \"%s\" (use %d)\n",
 		net,st->name,st->localcount);
-	free(net);
 	for (i=0; i<st->subnets->entries; i++) {
 	    net=subnet_to_string(st->subnets->list[i]);
 	    Message(c,"%s ",net);
-	    free(net);
 	}
 	if (i>0)
 	    Message(c,"-> host (use %d)\n",st->outcount);
@@ -798,8 +1058,7 @@ static void netlink_phase_hook(void *sst, uint32_t new_phase)
     /* All the networks serviced by the various tunnels should now
      * have been registered.  We build a routing table by sorting the
      * clients by priority.  */
-    st->routes=safe_malloc_ary(sizeof(*st->routes),st->n_clients,
-			       "netlink_phase_hook");
+    NEW_ARY(st->routes,st->n_clients);
     /* Fill the table */
     i=0;
     for (c=st->clients; c; c=c->next) {
@@ -828,12 +1087,16 @@ static void netlink_inst_set_mtu(void *sst, int32_t new_mtu)
 }
 
 static void netlink_inst_reg(void *sst, netlink_deliver_fn *deliver, 
-			     void *dst)
+			     void *dst, uint32_t *localmtu_r)
 {
     struct netlink_client *c=sst;
+    struct netlink *st=c->nst;
 
     c->deliver=deliver;
     c->dst=dst;
+
+    if (localmtu_r)
+	*localmtu_r=st->mtu;
 }
 
 static struct flagstr netlink_option_table[]={
@@ -893,7 +1156,7 @@ static closure_t *netlink_inst_create(struct netlink *st,
 	return NULL;
     }
 
-    c=safe_malloc(sizeof(*c),"netlink_inst_create");
+    NEW(c);
     c->cl.description=name;
     c->cl.type=CL_NETLINK;
     c->cl.apply=NULL;
@@ -986,6 +1249,8 @@ netlink_deliver_fn *netlink_init(struct netlink *st,
 	st->remote_networks=ipset_complement(empty);
 	ipset_free(empty);
     }
+    st->local_address=string_item_to_ipaddr(
+	dict_find_item(dict,"local-address", True, "netlink", loc),"netlink");
 
     sa=dict_find_item(dict,"secnet-address",False,"netlink",loc);
     ptpa=dict_find_item(dict,"ptp-address",False,"netlink",loc);
@@ -1009,7 +1274,7 @@ netlink_deliver_fn *netlink_init(struct netlink *st,
        though, and will make the route dump look complicated... */
     st->subnets=ipset_to_subnet_list(st->networks);
     st->mtu=dict_read_number(dict, "mtu", False, "netlink", loc, DEFAULT_MTU);
-    buffer_new(&st->icmp,ICMP_BUFSIZE);
+    buffer_new(&st->icmp,MAX(ICMP_BUFSIZE,st->mtu));
     st->outcount=0;
     st->localcount=0;
 
@@ -1060,7 +1325,7 @@ static list_t *null_apply(closure_t *self, struct cloc loc, dict_t *context,
     item_t *item;
     dict_t *dict;
 
-    st=safe_malloc(sizeof(*st),"null_apply");
+    NEW(st);
 
     item=list_elem(args,0);
     if (!item || item->type!=t_dict)
